/*
 * Amanda, The Advanced Maryland Automatic Network Disk Archiver
 * Copyright (c) 1991 University of Maryland
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of U.M. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  U.M. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * U.M. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL U.M.
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: James da Silva, Systems Design and Analysis Group
 *			   Computer Science Department
 *			   University of Maryland at College Park
 */
#include "amanda.h"
#include "stream.h"

static int bind_reserved(sock, addrp)
int sock;
struct sockaddr_in *addrp;
{
#   define FIRST_PORT	512
#   define NUM_PORTS  	(IPPORT_RESERVED-FIRST_PORT)
    int port, count;

    /* 
     * we pick a different starting point based on our pid to avoid always
     * picking the same reserved port twice.  this isn't really adequate,
     * but should be enough to mostly avoid our dumper race condition until
     * the next version of the dump protocol.
     */

    port = FIRST_PORT + (getpid() % NUM_PORTS);

    for(count = 0; count < NUM_PORTS; count++) {
	addrp->sin_port = htons(port);
	if(bind(sock, (struct sockaddr *)addrp, sizeof(struct sockaddr_in)) 
	   != -1) break;
	else if(errno != EADDRINUSE) {
	    /* something weird, probably EACCES */
	    return -1;
	}
	if(++port == IPPORT_RESERVED)
	    port = FIRST_PORT;
    }

    if(count >= NUM_PORTS) {
	/* all ports busy */
	errno = EAGAIN;
	return -1;
    }
    return 0;
}


int stream_server(portp)
int *portp;
{
    int server_socket, len;
    short port;
    struct sockaddr_in server;

    if((server_socket = socket(AF_INET, SOCK_STREAM, 0)) == -1)
	return -1;
    memset(&server, 0, sizeof(server));
    server.sin_family = AF_INET;
    server.sin_addr.s_addr = INADDR_ANY;

    if(geteuid() == 0) {
	if(bind_reserved(server_socket, &server) == -1) {
	    close(server_socket);
	    return -1;
	}
    }
    else {
	/* pick any available non-reserved port */
	server.sin_port = htons(0);

	if(bind(server_socket, (struct sockaddr *)&server, sizeof(server)) 
	   == -1) {
	    close(server_socket);
	    return -1;
	}
    }

    listen(server_socket, 0);

    /* find out what port was actually used */

    len = sizeof(server);
    if(getsockname(server_socket, (struct sockaddr *)&server, &len) == -1) {
	close(server_socket);
	return -1;
    }

    *portp = (int) ntohs(server.sin_port);
    return server_socket;
}

static void try_socksize(sock, which, size)
int sock, which, size;
{
    int origsize;

    origsize = size;
    /* keep trying, get as big a buffer as possible */
    while(size > 1024 &&
	  setsockopt(sock, SOL_SOCKET, which, &size, sizeof(int)) < 0)
	size -= 1024;
}


int stream_client(hostname, port, sendsize, recvsize)
char *hostname;
int port, sendsize, recvsize;
{
    int client_socket;
    struct sockaddr_in svaddr, claddr;
    struct hostent *hostp;

    if((hostp = gethostbyname(hostname)) == NULL)
	return -1;

    memset(&svaddr, 0, sizeof(svaddr));
    svaddr.sin_family = AF_INET;
    svaddr.sin_port = htons(port);
    memcpy(&svaddr.sin_addr, hostp->h_addr, hostp->h_length);

    if((client_socket = socket(AF_INET, SOCK_STREAM, 0)) == -1)
	return -1;

    if(geteuid() == 0) {	
	/* bind client side to reserved port before connect */
	memset(&claddr, 0, sizeof(claddr));
	claddr.sin_family = AF_INET;
	claddr.sin_addr.s_addr = INADDR_ANY;

	if(bind_reserved(client_socket, &claddr) == -1) {
	    close(client_socket);
	    return -1;
	}
    }

    if(connect(client_socket, (struct sockaddr *)&svaddr, sizeof(svaddr))
       == -1) {
	close(client_socket);
	return -1;
    }

    if(sendsize != DEFAULT_SIZE) 
	try_socksize(client_socket, SO_SNDBUF, sendsize);
    if(recvsize != DEFAULT_SIZE) 
	try_socksize(client_socket, SO_RCVBUF, recvsize);

    return client_socket;
}

/* don't care about these values */
static struct sockaddr_in addr;
static int addrlen;

int stream_accept(server_socket, timeout, sendsize, recvsize)
int server_socket, timeout, sendsize, recvsize;
{
    fd_set readset;
    struct timeval tv;
    int nfound, connected_socket;

    tv.tv_sec = timeout;
    tv.tv_usec = 0;
    FD_ZERO(&readset);
    FD_SET(server_socket, &readset);
    nfound = select(server_socket+1, &readset, NULL, NULL, &tv);
    if(nfound <= 0 || !FD_ISSET(server_socket, &readset))
	return -1;

    addrlen = sizeof(struct sockaddr);
    connected_socket = accept(server_socket,(struct sockaddr *)&addr,&addrlen);

    if(sendsize != DEFAULT_SIZE) 
	try_socksize(connected_socket, SO_SNDBUF, sendsize);
    if(recvsize != DEFAULT_SIZE) 
	try_socksize(connected_socket, SO_RCVBUF, recvsize);

    return connected_socket;
}
