/*
 * Amanda, The Advanced Maryland Automatic Network Disk Archiver
 * Copyright (c) 1991,1993 University of Maryland
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of U.M. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  U.M. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * U.M. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL U.M.
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: James da Silva, Systems Design and Analysis Group
 *			   Computer Science Department
 *			   University of Maryland at College Park
 */
/*
 * drv.c - controlling process for the Amanda backup system
 */
#include "amanda.h"
#include "clock.h"
#include "conffile.h"
#include "diskfile.h"
#include "logfile.h"
#include "statfs.h"

#include "driver.h"

#define LITTLE_DUMPERS 3

void driver_main()
{
    disklist_t *origqp;
    disk_t *diskp;
    fd_set selectset;
    int fd;
    dumper_t *dumper;
    char newdir[80];
    generic_fs_stats_t fs;

    startclock();
    FD_ZERO(&readset);

    if(read_conffile(CONFFILE_NAME))
	error("could not read amanda config file\n");

    construct_datestamp(datestamp);
    log(L_START,"date %s", datestamp);

    sprintf(taper_program, "%s/taper", LIBEXEC_DIR);
    sprintf(dumper_program, "%s/dumper", LIBEXEC_DIR);

    /* taper takes a while to get going; start it up right away */

    startup_tape_process();
    taper_cmd(START_TAPE, datestamp);

    /* start initializing; read in databases */

    if((origqp = read_diskfile(getconf_str(CNF_DISKFILE))) == NULL)
	error("could not read disklist file\n");

    /* set up any configuration-dependent variables */

    if(!force_parameters) {
	inparallel	= getconf_int(CNF_INPARALLEL);
	total_bandwidth = getconf_int(CNF_NETUSAGE);
	total_diskspace = getconf_int(CNF_DISKSIZE);
	diskdir		= getconf_str(CNF_DISKDIR);
	big_dumpers     = inparallel - LITTLE_DUMPERS;
	use_lffo	= 1;

	if(get_fs_stats(diskdir, &fs) == -1)
	    error("statfs %s: %s", diskdir, strerror(errno));
	if(fs.avail != -1 && fs.avail < total_diskspace) {
	    log(L_WARNING, 
		"WARNING: %d KB disksize requested, but only %d KB available.",
		total_diskspace, fs.avail);
	    total_diskspace = fs.avail;
	}
    }

    if(inparallel > MAX_DUMPERS) inparallel = MAX_DUMPERS;

    sprintf(newdir, "%s/%s", diskdir, datestamp);
    mkdir(newdir,0770);

    /* 
     * Read schedule from stdin.  Usually, this is a pipe from planner,
     * so the effect is that we wait here for the planner to
     * finish, but meanwhile the taper is rewinding the tape, reading
     * the label, checking it, writing a new label and all that jazz
     * in parallel with the planner.
     */

    waitq = *origqp;
    runq = read_schedule(&waitq);

    stoppedq.head = stoppedq.tail = NULL;
    tapeq.head = tapeq.tail = NULL;

    log(L_STATS, "startup time %s", walltime_str(curclock()));
printf("driver: start time %s inparallel %d bandwidth %d diskspace %d",
       walltime_str(curclock()), inparallel, total_bandwidth, total_diskspace);
printf(" dir %s datestamp %s driver: drain-ends tapeq %s big-dumpers %d\n",
       diskdir, datestamp,  use_lffo? "LFFO" : "FIFO", big_dumpers);
fflush(stdout);

    /* ok, planner is done, now lets see if the tape is ready */

    tok = getresult(taper);

    if(tok == TAPE_ERROR || tok == BOGUS) {
	/* no tape, go into degraded mode: dump to holding disk */
	degraded_mode = 1;
	force_incrementals(&runq);
	FD_CLR(taper,&readset);
    }

    startup_dump_processes();

    while(start_some_dumps(&runq) || some_dumps_in_progress()) {

short_dump_state();

	/* wait for results */

	memcpy(&selectset, &readset, sizeof(fd_set));
	if(select(maxfd+1, &selectset, NULL, NULL, NULL) == -1)
	    error("select: %s", strerror(errno));

	/* handle any results that have come in */

	for(fd = 0; fd <= maxfd; fd++) if(FD_ISSET(fd, &selectset)) {
	    if(fd == taper) handle_taper_result();
	    else handle_dumper_result(fd);
	}

    }

    /* handle any remaining dumps by dumping directly to tape, if possible */

    while(!empty(runq)) {
	diskp = dequeue_disk(&runq);
	if(!degraded_mode)
	    dump_to_tape(diskp);
	else 
	    log(L_FAIL, "%s %s %d [%s]",
		diskp->host->hostname, diskp->name, sched(diskp)->level,
		diskp->dtype->no_hold ?
		    "can't dump no-hold disk in degraded mode" :
		    "no more holding disk space");
    }

printf("driver: QUITTING time %s telling children to quit\n", 
       walltime_str(curclock()));
fflush(stdout);

    if(!degraded_mode)
	taper_cmd(QUIT, NULL);

    for(dumper = dmptable; dumper < dmptable + inparallel; dumper++)
	dumper_cmd(dumper, QUIT, NULL);

    /* wait for all to die */

    while(wait(NULL) != -1);

    if(!degraded_mode && rmdir(newdir))
	log(L_WARNING,"Could not rmdir(%s).  Check for cruft.", newdir);

printf("driver: FINISHED time %s\n", walltime_str(curclock()));
fflush(stdout);
    log(L_FINISH,"date %s time %s", datestamp, walltime_str(curclock()));
}

#define is_bigdumper(d) (((d)-dmptable) >= (inparallel-big_dumpers))

int start_some_dumps(rq)
disklist_t *rq;
{
    int total;
    disk_t *diskp, *nextp;
    dumper_t *dumper;

    total = 0;
#ifndef TOP_DOWN
    /*
     * A potential problem with starting from the bottom of the dump time
     * distribution is that a slave host will have both one of the shortest
     * and one of the longest disks, so starting its shortest disk first will
     * tie up the host and eliminate its longest disk from consideration the
     * first pass through.  This could cause a big delay in starting that long
     * disk, which could drag out the whole night's dumps.
     */
    for(dumper = dmptable; dumper < dmptable+inparallel; dumper++) {
#else
    /* 
     * While starting from the top of the dump time distribution solves the
     * above problem, this turns out to be a bad idea, because the big dumps
     * will almost certainly pack the holding disk completely, leaving no
     * room for even one small dump to start.  This ends up shutting out the
     * small-end dumpers completely (they stay idle).
     */
    for(dumper = dmptable+inparallel-1; dumper>=dmptable; dumper--) {
#endif
	if(dumper->busy) continue;
	/* found an idle dumper, now find a disk for it */
	if(is_bigdumper(dumper)) diskp = rq->tail;
	else diskp = rq->head;

	while(diskp) {
	    assert(diskp->host != NULL && sched(diskp) != NULL);

	    if(is_bigdumper(dumper)) nextp = diskp->prev;
	    else nextp = diskp->next;

	    if(!diskp->host->busy
	       && !diskp->dtype->no_hold
	       && sched(diskp)->est_size < free_space()	
	       && sched(diskp)->est_kps < free_kps()) {

		/* disk fits, dump it */
		allocate_bandwidth(sched(diskp)->est_kps);
		allocate_diskspace(sched(diskp)->est_size);
		dumper->busy = 1;		/* dumper is now busy */
		diskp->host->busy = 1;		/* host is now busy */

		dumper->dp = diskp;		/* link disk to dumper */
		sched(diskp)->dumper = dumper;
		total++;
		remove_disk(rq, diskp);		/* take it off the run queue */
		dumper_cmd(dumper, FILE_DUMP, diskp);
		break;
	    }
	    diskp = nextp;
	}
    }
    return total;
}

void force_incrementals(qp)
disklist_t *qp;
{
    disk_t *dp, *delp;

    dp = qp->head;
    while(dp != NULL) {
	if(sched(dp)->level == 0) {
	    if(sched(dp)->degr_level != -1) {
		sched(dp)->level = sched(dp)->degr_level;
		sched(dp)->est_size = sched(dp)->degr_size;
		sched(dp)->est_time = sched(dp)->degr_time;
		if(sched(dp)->est_time == 0) sched(dp)->est_kps = 10;
		else sched(dp)->est_kps = 
		    sched(dp)->est_size/sched(dp)->est_time;
		sprintf(sched(dp)->destname, "%s/%s/%s.%s.%d", 
			diskdir, datestamp, dp->host->hostname,
			dp->name, sched(dp)->level);
	    }
	    else {
		log(L_FAIL, "%s %s %d [can't switch to incremental dump]",
		    dp->host->hostname, dp->name, sched(dp)->level);
		delp = dp;
		dp = dp->next;
		remove_disk(qp, delp);
		continue;
	    }
	}
	dp = dp->next;
    }
}


void handle_taper_result()
{
    disk_t *dp;
    dumper_t *dumper;
    tok = getresult(taper);

    if(tok == DONE) {
	assert(argc >= 2);
	dp = serial2disk(argv[1]);
	free_serial(argv[1]);
	deallocate_diskspace(sched(dp)->est_size);
	unlink(sched(dp)->destname);
	printf("driver: finished-cmd time %s taper wrote %s:%s\n", 
	       walltime_str(curclock()),
	       dp->host->hostname, dp->name);
	fflush(stdout);

	if(empty(tapeq)) {
	    taper_busy = 0;
	    taper_disk = NULL;
	}
	else {
	    dp = dequeue_disk(&tapeq);
	    taper_disk = dp;
	    taper_cmd(FILE_WRITE, dp);
	}
	/*
	 * we need to restart some stopped dumps; without a good
	 * way to determine which ones to start, let them all compete
	 * for the remaining disk space.  Remember, having stopped
	 * processes indicates a failure in our estimation of sizes.
	 * Rather than have a complicated workaround to deal with
	 * stopped dumper processes more efficiently, we should
	 * work on getting better estimates to avoid the situation
	 * to begin with.
	 */
	while(!empty(stoppedq)) {
	    dp = dequeue_disk(&stoppedq);
	    dumper_cmd(sched(dp)->dumper, CONTINUE, NULL);
	}
    }
    else if(tok == TAPE_ERROR || tok == BOGUS) {
	dp = serial2disk(argv[1]);
	free_serial(argv[1]);
	printf("driver: finished-cmd time %s taper wrote %s:%s\n", 
	       walltime_str(curclock()),
	       dp->host->hostname, dp->name);
	fflush(stdout);

	/*
	 * Since we've gotten a tape error, we can't send anything more
	 * to the taper.  Go into degraded mode to try to get everthing
	 * onto disk.  Later, these dumps can be flushed to a new tape.
	 * The tape queue is zapped so that it appears empty in future
	 * checks.
	 */
	log(L_WARNING, "going into degraded mode because of tape error.");
	degraded_mode = 1;
	force_incrementals(&runq);
	tapeq.head = tapeq.tail = NULL;
	taper_busy = 0;
	taper_disk = NULL;
	FD_CLR(taper,&readset);
	close(taper);
    }
}


dumper_t *idle_dumper() 
{
    dumper_t *dumper;

    for(dumper = dmptable; dumper < dmptable+inparallel; dumper++)
	if(!dumper->busy) return dumper;

    return NULL;
}

int some_dumps_in_progress() 
{
    dumper_t *dumper;

    for(dumper = dmptable; dumper < dmptable+inparallel; dumper++)
	if(dumper->busy) return 1;

    return taper_busy;
}

int num_busy_dumpers() 
{
    dumper_t *dumper;
    int n;

    n = 0;
    for(dumper = dmptable; dumper < dmptable+inparallel; dumper++)
	if(dumper->busy) n += 1;

    return n;
}

static dumper_t *lookup_dumper(fd)
int fd;
{
    dumper_t *dumper;

    for(dumper = dmptable; dumper < dmptable+inparallel; dumper++)
	if(dumper->outfd == fd) return dumper;

    return NULL;
}

void construct_datestamp(buf)
char *buf;
{
    struct tm *tm;
    time_t timevar;

    timevar = time((time_t *)NULL);
    tm = localtime(&timevar);
    sprintf(buf, "%04d%02d%02d", tm->tm_year+1900, tm->tm_mon+1, tm->tm_mday);
}

int sort_by_time(a, b)
disk_t *a, *b;
{
    return sched(a)->est_time - sched(b)->est_time;
}

int sort_by_size_reversed(a, b)
disk_t *a, *b;
{
    return sched(b)->est_size - sched(a)->est_size;
}

void handle_dumper_result(fd)
int fd;
{
    dumper_t *dumper;
    disk_t *dp, *sdp;

    dumper = lookup_dumper(fd);
    tok = getresult(fd);

    if(tok == BOGUS) {
	/* either eof or garbage from dumper */
	log(L_WARNING, "dumper%d pid %d is messed up, ignoring it.\n", 
	    dumper-dmptable, dumper->pid);
	FD_CLR(fd,&readset);
	close(fd);
	dumper->busy = 1;	/* keep it busy so it isn't used again */
	if(dumper->dp) {
	    /* if it was dumping something, zap it and try again */
	    deallocate_bandwidth(sched(dumper->dp)->est_kps);
	    deallocate_diskspace(sched(dumper->dp)->est_size);
	    dumper->dp->host->busy = 0;		/* host is now idle */
	    unlink(sched(dumper->dp)->destname);
	    if(sched(dumper->dp)->attempted) {
		log(L_FAIL, "%s %s %d [dumper%d died]", 
		    dumper->dp->host->hostname, dumper->dp->name,
		    sched(dumper->dp)->level, dumper-dmptable);
	    }
	    else {
		log(L_WARNING, "dumper%d died while dumping %s:%s lev %d.", 
		    dumper-dmptable, 
		    dumper->dp->host->hostname, dumper->dp->name,
		    sched(dumper->dp)->level);
		sched(dumper->dp)->attempted++;
		enqueue_disk(&runq, dumper->dp);
	    }
	    dumper->dp = NULL;
	}
	return;
    }	

    sdp = serial2disk(argv[1]);	/* argv[1] always contains the serial number */

    assert(dumper->dp && sched(dumper->dp) && sdp == dumper->dp);


    switch(tok) {

    case DONE:
	free_serial(argv[1]);
	deallocate_bandwidth(sched(dumper->dp)->est_kps);
	dumper->busy = 0;
	dumper->dp->host->busy = 0;		/* host is now idle */
	printf("driver: finished-cmd time %s dumper%d dumped %s:%s\n", 
	       walltime_str(curclock()),
	       dumper-dmptable, dumper->dp->host->hostname, dumper->dp->name);
	fflush(stdout);

	if(taper_busy) {
	    if(use_lffo)insert_disk(&tapeq, dumper->dp, sort_by_size_reversed);
	    else enqueue_disk(&tapeq, dumper->dp);
	}
	else if(!degraded_mode) {
	    taper_disk = dumper->dp;
	    taper_busy = 1;
	    taper_cmd(FILE_WRITE, dumper->dp);
	}
	dumper->dp = NULL;
	break;

    case TRYAGAIN:
    case FATAL_TRYAGAIN:
	free_serial(argv[1]);
	deallocate_bandwidth(sched(dumper->dp)->est_kps);
	deallocate_diskspace(sched(dumper->dp)->est_size);
	dumper->busy = 0;
	dumper->dp->host->busy = 0;		/* host is now idle */
	unlink(sched(dumper->dp)->destname);
	
	if(sched(dumper->dp)->attempted) {
	    log(L_FAIL, "%s %s %d [could not connect to %s]", 
		dumper->dp->host->hostname, dumper->dp->name,
		sched(dumper->dp)->level, dumper->dp->host->hostname);
	}
	else {
	    sched(dumper->dp)->attempted++;
	    enqueue_disk(&runq, dumper->dp);
	}

	if(tok == FATAL_TRYAGAIN) {
	    /* dumper is confused, start another */
	    log(L_WARNING, "dumper%d (pid %d) confused, restarting it.", 
		dumper-dmptable, dumper->pid);
	    FD_CLR(fd,&readset);
	    close(fd);
	    startup_dump_process(dumper);
	}
	break;

    case FAILED:
	free_serial(argv[1]);
	deallocate_bandwidth(sched(dumper->dp)->est_kps);
	deallocate_diskspace(sched(dumper->dp)->est_size);
	dumper->busy = 0;
	dumper->dp->host->busy = 0;		/* host is now idle */
	unlink(sched(dumper->dp)->destname);

	/* no need to log this, dumper will do it */
	break;

    case NO_ROOM:
	if(empty(tapeq) && pending_aborts == 0) {
	    /* no disk space due to be freed */
	    dumper_cmd(dumper, ABORT, NULL);
	    pending_aborts++;
	    /* 
	     * if this is the only outstanding dump, it must be too big for
	     * the holding disk, so force it to go directly to tape on the
	     * next attempt.
	     */
	    if(num_busy_dumpers() <= 1)
		dumper->dp->dtype->no_hold = 1;
	}
	else {
	    /* !! bump up reserved space for this disk by some amount? */
	    enqueue_disk(&stoppedq, dumper->dp);
	}
	break;

    case ABORT_FINISHED:
	assert(pending_aborts);
	free_serial(argv[1]);
	deallocate_bandwidth(sched(dumper->dp)->est_kps);
	deallocate_diskspace(sched(dumper->dp)->est_size);
	unlink(sched(dumper->dp)->destname);
	sched(dumper->dp)->attempted++;
	enqueue_disk(&runq, dumper->dp);	/* we'll try again later */
	dumper->busy = 0;
	dumper->dp->host->busy = 0;		/* host is now idle */
	dumper->dp = NULL;
	pending_aborts--;
	while(!empty(stoppedq)) {
	    dp = dequeue_disk(&stoppedq);
	    dumper_cmd(sched(dp)->dumper, CONTINUE, NULL);
	}
	break;

    case BOGUS:
	/* either EOF or garbage from dumper.  Turn it off */
	break;

    default:
	assert(0);
    }
}

disklist_t read_schedule(waitqp)
disklist_t *waitqp;
{
    sched_t *sp;
    disk_t *dp;
    disklist_t rq;
    int rc, time, size, level, line, priority;
    int degr_level, degr_size, degr_time;
    char hostname[80], diskname[80];


    rq.head = rq.tail = NULL;

    /* read schedule from stdin */

    line = 0;
    while(!feof(stdin)) {
	line++;

	rc = scanf("%s %s %d %d %d %d %d %d %d\n", hostname, diskname, 
		   &priority, &level, &size, &time,
		   &degr_level, &degr_size, &degr_time);
	if(rc != 6 && rc != 9) {
	    error("schedule line %d: syntax error\n", line); 
	    continue;
	}

	dp = lookup_disk(hostname, diskname);
	if(dp == NULL) {
	    error("schedule line %d: %s:%s not in disk file\n", 
		  line, hostname, diskname);
	    continue;
	}
	
	sp = (sched_t *) alloc(sizeof(sched_t));
	sp->level    = level;
	sp->est_size = size;
	sp->est_time = time;
	if(rc < 9) sp->degr_level = -1;
	else {
	    sp->degr_level = degr_level;
	    sp->degr_size = degr_size;
	    sp->degr_time = degr_time;
	}
	if(time == 0) sp->est_kps = 10;
	else sp->est_kps  = size/time;
	sp->attempted = 0;
	sp->act_size = 0;
	sp->dumper  = NULL;
	sprintf(sp->destname, "%s/%s/%s.%s.%d", 
		diskdir, datestamp, hostname, diskname, level);
	dp->up = (char *) sp;
	remove_disk(waitqp, dp);
	insert_disk(&rq, dp, sort_by_time);
    }
    return rq;
}

int free_kps() 
{
    return total_bandwidth-allocated_bandwidth;
}

void allocate_bandwidth(kps) 
int kps;
{
    allocated_bandwidth += kps;
}

void deallocate_bandwidth(kps)
int kps;
{
    assert(kps <= allocated_bandwidth);
    allocated_bandwidth -= kps;
}

int free_space() 
{
    return total_diskspace-allocated_diskspace;
}

void allocate_diskspace(size) 
int size;
{
    assert(size < free_space());
    allocated_diskspace += size;
}

void deallocate_diskspace(size)
int size;
{
    assert(size <= allocated_diskspace);
    allocated_diskspace -= size;
}

int dump_to_tape(dp)
disk_t *dp;
{
    dumper_t *dumper;
    int failed = 0;

    inside_dump_to_tape = 1;	/* for simulator */

printf("driver: dumping %s:%s directly to tape\n",
       dp->host->hostname, dp->name);

    /* tell the taper to read from a port number of its choice */

    taper_cmd(PORT_WRITE, dp);
    tok = getresult(taper);
    if(tok != PORT) return 1;
    strcpy(sched(dp)->destname, argv[1]);	/* copy port number */

    /* pick a dumper, then tell it to dump to a port */

    dumper = idle_dumper();
    dumper_cmd(dumper, PORT_DUMP, dp);

    /* update statistics & print state */

    taper_busy = dumper->busy = dp->host->busy = 1;
    allocate_bandwidth(sched(dp)->est_kps);

    short_dump_state();

    /* wait for result from dumper */

    tok = getresult(dumper->outfd);
    free_serial(argv[1]);	/* reap the serial number */
    switch(tok) {
    case DONE:
	/* everthing went fine */
	break;
    case NO_ROOM:
	dumper_cmd(dumper, ABORT, dp);
	tok = getresult(dumper->outfd);
	free_serial(argv[1]);	/* reap the serial number */
	assert(tok == ABORT_FINISHED);
    case TRYAGAIN:
    default:
	/* dump failed, but we must still reap result from taper */
	failed = 1;
    }

    tok = getresult(taper);
    free_serial(argv[1]);	/* reap the serial number */
    if(tok != DONE) failed = 1;

    /* reset statistics & return */

    taper_busy = dumper->busy = dp->host->busy = 0;
    deallocate_bandwidth(sched(dp)->est_kps);

    inside_dump_to_tape = 0;
    return failed;
}

int queue_length(q)
disklist_t q;
{
    disk_t *p;
    int len;

    for(len = 0, p = q.head; p != NULL; len++, p = p->next);
    return len;
}


void short_dump_state()
{
    int i, nidle;

    printf("driver: state time %s ", walltime_str(curclock()));
    printf("free kps: %d space: %d taper: ", free_kps(), free_space());
    if(degraded_mode) printf("DOWN");
    else if(!taper_busy) printf("idle");
    else printf("writing");
    nidle = 0;
    for(i = 0; i < inparallel; i++) if(!dmptable[i].busy) nidle++;
    printf(" idle dumpers: %d", nidle);
    printf(" qlen tapeq: %d", queue_length(tapeq));
    printf(" runq: %d",       queue_length(runq));
    printf(" stoppedq: %d\n", queue_length(stoppedq));
    fflush(stdout);
}

void dump_state(str)
char *str;
{
    int i;
    disk_t *dp;

    printf("================\n");
    printf("driver state at time %s: %s\n", walltime_str(curclock()), str);
    printf("free kps: %d, space: %d\n", free_kps(), free_space());
    if(degraded_mode) printf("taper: DOWN\n");
    else if(!taper_busy) printf("taper: idle\n");
    else printf("taper: writing %s:%s.%d est size %d\n", 
		taper_disk->host->hostname, taper_disk->name, 
		sched(taper_disk)->level,
		sched(taper_disk)->est_size);
    for(i = 0; i < inparallel; i++) {
	dp = dmptable[i].dp;
	if(!dmptable[i].busy) printf("dumper%02d: idle\n", i);
	else printf("dumper%02d: dumping %s:%s.%d est kps %d size %d\n", i,
		    dp->host->hostname, dp->name, sched(dp)->level,
		    sched(dp)->est_kps, sched(dp)->est_size);
    }
    dump_queue("TAPE", tapeq, 5, stdout);
    dump_queue("STOP", stoppedq, 5, stdout);
    dump_queue("RUN ", runq, 5, stdout);
    printf("================\n");
    fflush(stdout);
}
